/*
        This file is part of DISSECT.

        DISSECT is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.

        DISSECT is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with DISSECT.  If not, see <http://www.gnu.org/licenses/>.
*/



/**
 * Created by Graham on 28/02/14.
 */

package dissect;

import beast.core.Citation;
import beast.core.Description;
import beast.core.Input;
import beast.core.parameter.RealParameter;
import beast.evolution.speciation.YuleModel;
import beast.evolution.tree.Node;
import beast.evolution.tree.TreeInterface;

import java.util.List;



@Description("Birth Death Collapse model. <br/>")

@Citation(value = "Jones and Oxelman. " +
        "DISSECT: an assignment-free Bayesian discovery method for species delimitation under the multispecies coalescent. " +
        "http://biorxiv.org/content/early/2014/03/03/003178. doi: 10.1101/003178.",
        firstAuthorSurname = "Jones")

// It would seem more natural to extend SpeciesTreeDistribution, but its
// calculateTreeLogLikelihood() is package private.
public class BirthDeathCollapseModel extends YuleModel {
    public static final Input<Double> collapseHeight =
            new Input<Double>("collapseHeight", "collapse height value, epsilon in birth/death/collapse model", Input.Validate.REQUIRED);
    public Input<RealParameter> relativeDeathRate =
            new Input<RealParameter>("relativeDeathRate", "relative death rate parameter, mu/lambda in birth/death/collapse model", Input.Validate.REQUIRED);
    public Input<RealParameter> collapseWeight =
            new Input<RealParameter>("collapseWeight", "collapse weight parameter, w in birth/death/collapse model", Input.Validate.REQUIRED);
    public Input<RealParameter> originHeight =
            new Input<RealParameter>("originHeight", "origin height parameter, in birth/death/collapse model", Input.Validate.REQUIRED);


    @Override
    public void initAndValidate() throws Exception {
        super.initAndValidate();
        if (conditionalOnRoot) {
            System.err.println("WARNING: dissect.BirthDeathCollapseModel cannot handle conditionalOnRoot=true.");
        }
        if ((collapseHeight.get() < 1e-30) || (collapseHeight.get() > 1e10)) {
            System.err.println("WARNING: dissect.BirthDeathCollapseModel: bad collapseHeight value");
        }
        if (collapseWeight.get().lowerValueInput.get() < 0  ||
                collapseWeight.get().upperValueInput.get() > 1.0 ) {
            System.err.println("WARNING: dissect.BirthDeathCollapseModel: bad collapseWeight limits");
        }
        if (relativeDeathRate.get().lowerValueInput.get() < 0  ||
                relativeDeathRate.get().upperValueInput.get() > 1.0 ) {
            System.err.println("WARNING: dissect.BirthDeathCollapseModel: bad relativeDeathRate limits");
        }
        // Initialise originHeight to be a bit bigger than root height
        TreeInterface tree = treeInput.get();
        Double [] initRHarray = new Double[1];
        initRHarray[0] = 1.05 * tree.getRoot().getHeight();
        RealParameter initRH = new RealParameter(initRHarray);
        originHeight.get().assignFromWithoutID(initRH);
    }

    @Override
    public double calculateTreeLogLikelihood(final TreeInterface tree) {
        double logpt = 0.0;
        int ntips = tree.getLeafNodeCount();
        double alpha = birthDiffRateParameterInput.get().getValue();
        double beta = relativeDeathRate.get().getValue();
        double tor = originHeight.get().getValue();
        double w = collapseWeight.get().getValue();

        double rooth = tree.getRoot().getHeight();
        if (rooth > tor) {
            return Double.NEGATIVE_INFINITY;
        }

        logpt += originHeightLogLikelihood(tor, alpha, beta, w, ntips);

        List<Node> internalNodes = tree.getInternalNodes();
        for (Node node : internalNodes) {
            final double height = node.getHeight();
            double usualpn = nodeHeightLikelihood(height, tor, alpha, beta);
            double collapsedpn = belowCollapseHeight(height) ? (1.0 / collapseHeight.get()) : 0.0;
            logpt += Math.log((1.0 - w) * usualpn + w * collapsedpn);
        }

        return logpt;
    }





    public double getCollapseHeight() {
        return collapseHeight.get();
    }

    // provided to help avoid inconsistent treatment of h == collapseHeight
    public static boolean belowCollapseHeight(double h) {
        return (h < collapseHeight.get());
    }





    private double originHeightLogLikelihood(double t, double a, double b, double w, int n) {
        double E = Math.exp(-a * t);
        double B = (1 - E) / (1-b*E);
        double z = 0.0;
        z += Math.log(a);
        z += Math.log(1 - b);
        z -= a * t;
        z -= 2 * Math.log(1 - b * E);
        z += (n-2) * Math.log(w + (1 - w) * B);
        z +=  Math.log(w + n * (1 - w) * B);
        return z;
    }


    private double nodeHeightLikelihood(double s, double t, double a, double b) {
        double Es = Math.exp(-a * s);
        double Et = Math.exp(-a * t);
        double z = 0.0;
        if (s < t) {
            z = a;
            z *= (1 - b);
            z *= Es;
            z /= (1 - b * Es) * (1 - b * Es);
            z *= (1 - b * Et);
            z /= (1 - Et);
        }
        return z;
    }

    @Override
    protected boolean requiresRecalculation() {
        return super.requiresRecalculation() ||
                relativeDeathRate.get().somethingIsDirty() ||
                collapseWeight.get().somethingIsDirty() ||
                originHeight.get().somethingIsDirty();
    }
}

